class SearchResult:
    """SearchResult contains the results for a search call."""

    def __init__(self, results):
        self.results = results

    @property
    def best(self):
        return self.results[0]



import numpy as np

import keras_genetic


class Breeder:
    """Breeder is responsible for creating offspring."""

    def __init__(self, initializer=None):
        if initializer is None:
            initializer = keras_genetic.initializers.random_normal.RandomNormal()
        self.initializer = initializer

    def offspring(self, parents):
        """`offspring()` creates a new offspring from a mother and father model.

        Args:
            parents: list of keras_genetic.Individual.
        Returns:
            a new keras_genetic.Individual representing the offspring.
        """
        raise NotImplementedError(
            "`offspring()` must be implemented on the breeder class."
        )

    def population_from_parents(self, parents, population_size):
        result = []
        for _ in range(population_size):
            result.append(self.offspring(parents))
        return result

    def fully_random_weight_set(self, mother):
        offspring_weights = []

        # each 'mother.weight', 'father.weight' we have a vector of weights
        # we must traverse the entire array and sample a random one for each
        for m in mother.weights:
            shape = m.shape

            # easier to traverse them as flattened arrays
            m = m.flatten()
            result = []
            for i in range(m.shape[0]):
                result.append(self.initializer())

            offspring_weights.append(np.array(result).reshape(shape))

        return keras_genetic.Individual(offspring_weights, model=mother.model)



import random

import numpy as np

from keras_genetic import core
from keras_genetic.breeder.breeder import Breeder


class MutationBreeder(Breeder):
    """MutationBreeder randomly mutates features on individuals."""

    def __init__(
        self,
        keep_probability=0.9,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.keep_probability = keep_probability

    def offspring(self, parents):
        mother = random.choice(parents)
        offspring_weights = []

        # each 'mother.weight', 'father.weight' we have a vector of weights
        # we must traverse the entire array and sample a random one for each
        for w in mother.weights:
            shape = w.shape
            w = w.flatten()

            result = []
            for i in range(w.shape[0]):
                if random.uniform(0, 1) > self.keep_probability:
                    result.append(self.initializer())
                    continue
                result.append(w[i])

            offspring_weights.append(np.array(result).reshape(shape))

        return core.Individual(offspring_weights, model=mother.model)



from keras_genetic.breeder.breeder import Breeder
from keras_genetic.breeder.mutation_breeder import MutationBreeder
from keras_genetic.breeder.n_parent_mutation_breeder import NParentMutationBreeder



class Individual:
    """Individual represents a single individual.

    Args:
        weights: array of np.array weights.
        model: keras model to load the weights into.
        score: score the individual received.

    Attributes:
        weights: array of np.arrays representing weights of the neural network.
        score: score the individual received during training.
    """

    def __init__(self, weights, model, score=None):
        self.weights = weights
        self.model = model
        self.score = score

    def load_model(self):
        """load_model() prepares the keras model for use.

        Usage:
        ```python
        model = individual.load_model()
        y_test = model.predict(x_test)
        ```
        """
        self.model.set_weights(self.weights)
        return self.model

    def __gt__(self, other):
        return self.score > other.score

    def __lt__(self, other):
        return self.score < other.score



import numpy as np

from keras_genetic.initializers.initializer import Initializer


class RandomNormal(Initializer):
    def __init__(self, mean=0, std=10):
        self.mean = mean
        self.std = std

    def __call__(self):
        return np.random.normal(self.mean, self.std)



import keras_genetic


class CompletelyRandomBreeder(keras_genetic.Breeder):
    """CompletelyRandomBreeder creates fully random weight sets."""

    def offspring(self, parents):
        mother = parents[0]
        return self.fully_random_weight_set(mother)



"""search is the primary entrypoint to keras-genetic."""

import tqdm

from keras_genetic import core
from keras_genetic.search.result import SearchResult


def search(
    model,
    generations,
    population_size,
    n_parents_from_population,
    evaluator,
    breeder,
    return_best=1,
    callbacks=(),
):
    """`search()` is the primary entrypoint to keras-genetic.

    `search()` takes in a keras.Model, a breeder, and an evaluator.
    Using these three parameters, it searches the weight space and returns the
    n-best weights according to the `return_best` parameter.

    Args:
        model: the keras.Model to run the seach over.
        generations: number of generations to run the genetic algorithm for.
        population_size: size of the population.
        n_parents_from_population: number of parents to keep from population.
        evaluator: `keras_genetic.Evaluator` instance, or callable.  Must take
            a model instance and return a number indicating the fitness score
            of the individual.  If you wish to minimize a loss, simply
            invert the result of your loss function.
        breeder: `keras_genetic.Breeder` instance (or callable) used to produce
            new offspring
        return_best: number of `keras_genetic.Individual` to return.
    Returns:
        `keras_genetic.search.SearchResult` object containing all of the search results
    """
    search_manager = _SearchManager(
        generations,
        population_size,
        n_parents_from_population,
        breeder,
        evaluator,
        return_best,
        callbacks,
    )
    return search_manager.run(model)


class _SearchManager:
    def __init__(
        self,
        generations,
        population_size,
        n_parents_from_population,
        breeder,
        evaluator,
        return_best,
        callbacks,
    ):
        self.generations = generations
        self.population_size = population_size
        self.n_parents_from_population = n_parents_from_population
        self.breeder = breeder
        self.evaluator = evaluator
        self.return_best = return_best
        self.callbacks = callbacks

    def initial_generation(self, model, dummy_individual):
        return [
            self.breeder.fully_random_weight_set(dummy_individual)
            for _ in range(self.population_size)
        ]

    def run_generation(self, population, parents, keep):
        for individual in population:
            individual.score = self.evaluator(individual)

        result_population = sorted(population + parents, reverse=True)
        return result_population[:keep]

    def run(self, model):
        initial_parent = core.Individual(model.get_weights(), model)
        population = self.initial_generation(model, initial_parent)
        parents = []

        for g in tqdm.tqdm(range(self.generations)):

            for callback in self.callbacks:
                callback.on_generation_begin(g, SearchResult(population))

            parents = self.run_generation(
                population, parents, keep=self.n_parents_from_population
            )
            population = self.breeder.population_from_parents(
                parents, self.population_size
            )

            for callback in self.callbacks:
                callback.on_generation_end(g, SearchResult(parents))

        final = self.run_generation(population, [], keep=self.return_best)
        return SearchResult(final)



import random

import numpy as np

from keras_genetic import core
from keras_genetic.breeder.breeder import Breeder


class NParentMutationBreeder(Breeder):
    """NParentMutationBreeder randomly mutates features on individuals."""

    def __init__(
        self,
        n,
        keep_probability=0.9,
        **kwargs,
    ):
        super().__init__(**kwargs)
        if n < 1:
            raise ValueError(
                f"NParentMutationBreeder.__init__() received n={n}.  Want n > 1"
            )
        self.n = n
        self.keep_probability = keep_probability

    def offspring(self, parents):
        parents = random.sample(parents, self.n)
        offspring_weights = []

        # each 'mother.weight', 'father.weight' we have a vector of weights
        # we must traverse the entire array and sample a random one for each
        for parent_weights in zip(parents):
            shape = parent_weights[0].shape
            parent_weights = [w.flatten() for w in parent_weights]
            # easier to traverse them as flattened arrays
            result = []
            for i in range(parent_weights[0].shape[0]):
                if random.uniform(0, 1) > self.keep_probability:
                    result.append(self.initializer())
                    continue
                choice = random.choice(parent_weights)
                result.append(choice[i])

            offspring_weights.append(np.array(result).reshape(shape))

        return core.Individual(offspring_weights, model=parents[0].model)



from keras_genetic.breeder import Breeder
from keras_genetic.core.individual import Individual
from keras_genetic.initializers import Initializer
from keras_genetic.search import search

__version__ = "0.0.4"
